// Flame.cpp: Implementierung der Klasse Flame
//
///////////////////////////////////////////////////////////////////////

/*

  This class is heavily based on code written by:
	
	Copyright (c) 2002 Yury Uralsky
	Yury Uralsky (carlos@blaze.ru, yuryu@kaluga.ru, ICQ: 42069218)


*/

//#include <GLH/glh_nveb.h>
//#include <GLH/glh_extensions.h>
//#include <GLH/glh_linear.h>
//#include <GLH/glh_obs.h>
//#include <GLH/glh_glut.h>
//#include <GLH/load_cubemap.h>

//#include <array_texture.h>
//#include <NV/nv_png.h>
//#include <NV/nv_dds.h>
//#include <shared/data_path.h>
//#include <shared/noise.h>

#include "Cg_Shader.h"
#include "Fire.h"

//using namespace glh;
//using namespace glh::ns_float;

glh::display_list fire_quads;
//glh::tex_object_2D fire_tex;
glh::tex_object_rectangle gradients_tex, permutations_tex;




Fire::Fire()
{
	bbox_enabled = false;
	outlines_enabled = false; 
	fire_enabled = true;
	animation = true;

	frames = 0;

	slices_count = 10;

	time = 0, 
	speed = 0.01f;
	turb_scale = 2.0f;

}

Fire::~Fire()
{
	clear();


}

void Fire::create_quads_list()
{
	//glNewList(fireQuads,GL_COMPILE);
	fire_quads.new_list(GL_COMPILE);
	glBegin(GL_QUADS);
	for (int i = 1; i<slices_count+1; i++)
	{
		float z = (float)i / (slices_count+1);
		glVertex3f(0, 0, z);
		glVertex3f(1, 0, z);
		glVertex3f(1, 1, z);
		glVertex3f(0, 1, z);
	}
	glEnd();
	fire_quads.end_list();
	//glEndList();
}

void Fire::create_textures()
{
	float g[B + B + 2][4];	// 4d gradient vectors
	float p[B + B + 2];		// permutations

	srand(GetTickCount());

	int i = 0, j = 0;

	while (i < B)
	{
		glh::ns_float::vec4 v(2 * frand() - 1, 2 * frand() - 1, 2 * frand() - 1, 2 * frand() - 1);

		if (v.normalize() < 1)
		{
			v.get_value(g[i][0], g[i][1], g[i][2], g[i][3]);
			++i;
		}
	}

	// Generate permutations

	for (i = 0; i<B; i++) p[i] = i;

	while (--i) {
		int k = p[i];
		int j;
		p[i] = p[j = rand() % B];
		p[j] = k;
	}

	for (i = 0; i<B + 2; i++) {
		p[B + i] = p[i];
		for (j = 0; j<4; j++) g[B + i][j] = g[i][j];
	}


	// load gradients texture
	glPushAttrib(GL_TEXTURE_BIT);
	gradients_tex.bind();
	//glGenTextures(1, &gradientHandle);
    //glBindTexture(GL_TEXTURE_RECTANGLE_ARB, gradientHandle);
	//glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_FLOAT_RGBA32_NV, B + B + 2, 1, 0, GL_RGBA, GL_FLOAT, g);
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA32F_ARB, B + B + 2, 1, 0, GL_RGBA, GL_FLOAT, g);

	// load permutations texture
	permutations_tex.bind();
	//glGenTextures(1, &permutationHandle);
    //glBindTexture(GL_TEXTURE_RECTANGLE_ARB, permutationHandle);
	//glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_FLOAT_R32_NV, B + B + 2, 1, 0, GL_RED, GL_FLOAT, p);
	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA32F_ARB, B + B + 2, 1, 0, GL_RED, GL_FLOAT, p);
	
	glPopAttrib();
	// load the fire texture

	//fire_tex.bind();
	//array2<vec3ub> fire;
	//read_png_rgb("firetex.png", fire);
	//make_rgb_texture(fire, false);
	//fire_tex.parameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	//fire_tex.parameter(GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//fire_tex.parameter(GL_TEXTURE_WRAP_S, GL_CLAMP);
	//fire_tex.parameter(GL_TEXTURE_WRAP_T, GL_CLAMP);

	Image* flameTexture = FlipImage(OpenImage("flame.png",PF_R8G8B8),CA_X);
	Array3<unsigned char> flameData(flameTexture->getWidth(), flameTexture->getHeight(),3,(unsigned char*)flameTexture->getPixels());
	glPushAttrib(GL_TEXTURE_BIT);

    // Generate the texture
    glGenTextures(1, &fireHandle);
    glBindTexture(GL_TEXTURE_2D, fireHandle);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_CLAMP);


    gluBuild2DMipmaps(GL_TEXTURE_2D,
        GL_RGB8,
        flameData.getWidth(),
        flameData.getHeight(),
        GL_RGB,
        GL_UNSIGNED_BYTE,
        flameData.getData());

    glPopAttrib();

}

void Fire::clear()
{
	// destroy the texture references
	glDeleteTextures(1,&fireHandle);
	glDeleteTextures(1,&permutationHandle);
	glDeleteTextures(1,&gradientHandle);


	if(cgContext != NULL)
	{
		//cgFreeContext(Context);
		cgDestroyContext(cgContext);
	}
	
}

bool Fire::init()
{

	// define callback function for Cg errors
	cgSetErrorCallback(cgErrorCallback);
	// create a Cg context which will hold our shader programs
	cgContext = cgCreateContext();

	// get latest vertex shader profile supported by the installed HW
	cgVertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
	cgGLSetOptimalOptions(cgVertexProfile);
	if (cgVertexProfile == CG_PROFILE_UNKNOWN)
	{
		cout << "Graphics card not supported; could not find a suitable vertex shader profile.\n";
	}

	// get latest fragment shader profile supported by the installed HW
	cgFragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
	cgGLSetOptimalOptions(cgFragmentProfile);
	if (cgFragmentProfile == CG_PROFILE_UNKNOWN)
	{
		cout << "Graphics card not supported; could not find a suitable fragment shader profile.\n";
	}

	cout << "-Fire vertex program loading...-\n";
	// create actual Cg program from file; last two parameters are entry point and arguments
	cgVertexProgram = cgCreateProgramFromFile(cgContext, CG_SOURCE, "cg_fireVP.cg", cgVertexProfile, 0, 0);
	if(!cgVertexProgram)
	{
		cout << "Could not read and create vertex shader program.\n";
	}
    cgGLLoadProgram(cgVertexProgram);
	cout << "-Fire vertex program loaded.-\n";

	cout << "-Fire fragment program loading...-\n";
	cgFragmentProgram = cgCreateProgramFromFile(cgContext, CG_SOURCE, "cg_fireFP.cg", cgFragmentProfile, 0, 0);
	if(!cgFragmentProgram)
	{
		cout << "Could not read and create fragment shader program.\n";
	}
	cgGLLoadProgram(cgFragmentProgram);
	cout << "-Fire fragment program loaded.-\n";
	
	Time = cgGetNamedParameter(cgVertexProgram, "time");
	if(Time == NULL)
	{
		cerr << "Cannot get a handle to the time variable in the fire cg vertex program" << endl;
		return false;
	}

	ModelView = cgGetNamedParameter(cgVertexProgram, "modelview");
	if(ModelView == NULL)
	{
		cerr << "Cannot get a handle to the model view matrix in the fire cg vertex program" << endl;
		return false;
	}

	InverseModelView = cgGetNamedParameter(cgVertexProgram, "imodelview");
	if(InverseModelView == NULL)
	{
		cerr << "Cannot get a handle to the inverse model view matrix in the fire cg vertex program" << endl;
		return false;
	}

	Proj = cgGetNamedParameter(cgVertexProgram, "proj");
	if(Proj == NULL)
	{
		cerr << "Cannot get a handle to the projection matrix in the fire cg vertex program" << endl;
		return false;
	}

	Turb_Scale = cgGetNamedParameter(cgFragmentProgram, "turb_scale");
	if(Turb_Scale == NULL)
	{
		cerr << "Cannot get a handle to the turbulence scale in the fire cg vertex program" << endl;
		return false;
	}
	
	/*PBind = cgGetNamedParameter(cgFragmentProgram, "p");
	if(PBind == NULL)
	{
		cerr << "Cannot get a handle to the perturbation texture in the fire cg fragment program" << endl;
		return false;
	}*/
	
	/*GBind = cgGetNamedParameter(cgFragmentProgram, "g");
	if(GBind == NULL)
	{
		cerr << "Cannot get a handle to the gradient texture in the fire cg fragment program" << endl;
		return false;
	}*/
	
	/*FireTexBind = cgGetNamedParameter(cgFragmentProgram, "firetex");
	if(FireTexBind == NULL)
	{
		cerr << "Cannot get a handle to the fire texture in the fire cg fragment program" << endl;
		return false;
	}*/

	cout << "Loading fire textures..." << endl;
	create_textures();
	cout << "...fire textures loaded." << endl;

	create_quads_list();

	return true;
	
}

/*
void keyboard(unsigned char k, int x, int y)
{
	if (k == 27 || k == 'q')
	{
		shutdown();
		exit(0);
	}

	switch (k)
	{
	case 'a':
		animation = !animation;
		break;
	case 's':
		speed -= 0.0025f;
		if (speed < 0.0025f) speed = 0.0025f;
		break;
	case 'S':
		speed += 0.0025f;
		if (speed > 0.5f) speed = 0.5f;
		break;
	case 'f':
		fire_enabled = !fire_enabled;
		break;
	case 'b':
		bbox_enabled = !bbox_enabled;
		break;
	case 'o':
		outlines_enabled = !outlines_enabled;
		break;
	case '+':
		++slices_count;
		if (slices_count > 100) slices_count = 100;
		create_quads_list();
		break;
	case '-':
		--slices_count;
		if (slices_count < 5) slices_count = 5;
		create_quads_list();
		break;
	case 't':
		turb_scale -= 0.25f;
		if (turb_scale < 0) turb_scale = 0;
		break;
	case 'T':
		turb_scale += 0.25f;
		if (turb_scale > 5.0f) turb_scale = 5.0f;
		break;

	}

	glutPostRedisplay();
}
*/

/*
void idle()
{
	if (animation)
	{
		time += speed;

		glh::ns_float::vec4 diffuse(0, 0, 0, 1);
		float k = frand();
		diffuse[0] = 0.5f + 0.3f * k;
		diffuse[1] = 0.1f + 0.2f * k;
		glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse.v);
	}

	glutPostRedisplay();
}
*/




void Fire::draw_outlines()
{
	cgGLEnableProfile(cgVertexProfile);
	cgGLBindProgram(cgVertexProgram);

	for (int i = 0; i<6; i++) glEnable(GL_CLIP_PLANE0 + i);

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glColor3f(0, 0, 1);
	fire_quads.call_list();
	//glCallList(fireQuads);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	for (i = 0; i<6; i++) glDisable(GL_CLIP_PLANE0 + i);

	cgGLDisableProfile(cgVertexProfile);

}

void Fire::draw_fire(float update_time)
{
	// Set some global render states

	//time = update_time;

	static const glh::ns_float::vec4 ambient(0, 0, 0, 1);

	glPushAttrib(GL_ACCUM_BUFFER_BIT|GL_CURRENT_BIT|GL_LIGHTING_BIT|GL_TEXTURE_BIT|GL_DEPTH_BUFFER_BIT|GL_ENABLE_BIT|GL_TRANSFORM_BIT|GL_VIEWPORT_BIT|GL_COLOR_BUFFER_BIT);
	glClearColor(0, 0, 0, 0);
	glBlendFunc(GL_ONE, GL_ONE);
	glEnable(GL_LIGHT0);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient.v);
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_DEPTH_TEST);

	// enable the shaders for the fire effect
	cgGLEnableProfile(cgVertexProfile);
	cgGLEnableProfile(cgFragmentProfile);
	cgGLBindProgram(cgVertexProgram);
	cgGLBindProgram(cgFragmentProgram);
	
	for (int i = 0; i<6; i++) glEnable(GL_CLIP_PLANE0 + i);

	// assign the model view matrix
	cgGLSetStateMatrixParameter(ModelView, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY);

	// assign the inverse model view matrix
	cgGLSetStateMatrixParameter(InverseModelView, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_INVERSE);

	// assign the projection matrix
	cgGLSetStateMatrixParameter(Proj, CG_GL_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);

	// assign the time
	//cgGLSetParameter4f(Time, time, 0.0f, 0.0f, 0.0f);
	cgGLSetParameter1f(Time, time);
	// assign the turbulance scale
	//cgGLSetParameter4f(Turb_Scale, turb_scale, 0.0f, 0.0f, 0.0f);
	cgGLSetParameter1f(Turb_Scale, turb_scale);

	// new: set COLOR0 instead of DIFFUSE
	float k = frand();
	//glColor3f(0.5f + 0.3f * k, 0.1f + 0.2f * k, 0.0f);

	// activate the textures
	
	//cgGLActiveTexture(GBindIter);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	//glEnable(GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_RECTANGLE_ARB);
	//glBindTexture(GL_TEXTURE_2D, gradientHandle);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, gradientHandle);
	//gradients_tex.bind();

	//cgGLActiveTexture(PBindIter);
	glActiveTextureARB(GL_TEXTURE2_ARB);
	//glEnable(GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_RECTANGLE_ARB);
	glBindTexture(GL_TEXTURE_2D, permutationHandle);
	//permutations_tex.bind();

	glActiveTextureARB(GL_TEXTURE4_ARB);
	glEnable(GL_TEXTURE_2D);	
	glBindTexture(GL_TEXTURE_2D, fireHandle);
	//fire_tex.bind();

	glEnable(GL_BLEND);
	fire_quads.call_list();
	glDisable(GL_BLEND);

	for (i = 0; i<6; i++) glDisable(GL_CLIP_PLANE0 + i);
	
	// disable the vertex and fragment profiles
	cgGLDisableProfile(cgVertexProfile);
	cgGLDisableProfile(cgFragmentProfile);

	glPopAttrib();

}


/*
void display()
{

	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glScalef(4, 10, 4);	// scale fire volume

	if (bbox_enabled)
	{
		glColor3f(1, 1, 1);
		glutWireCube(1);
	}

	glTranslatef(-0.5f, -0.5f, -0.5f);

	if (outlines_enabled)
		draw_outlines();	// draw slice outlines

	if (fire_enabled)
		draw_fire();		// draw volumetric fire

	glPopMatrix();

}
*/