// OBJLoader.cpp: Implementierung der Klasse OBJLoader
//
///////////////////////////////////////////////////////////////////////

#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include"OBJLoader.h"


OBJLoader::OBJLoader()
{
	totalVertices = 0;
	totalFaces = 0;
	vertices = 0;
	polyList = 0;
}


OBJLoader::~OBJLoader()
{
	clear();
}


bool OBJLoader::LoadObject(char *filename)
{
	FILE *file;
	char buffer[256] = {0};

	if(filename == NULL) return false;

	if(!(file = fopen(filename, "r")))
		return false;

	//clear old data
	clear();


    /*
     * First we check the how many vertices and faces there are in the model,
     * for this we scan the whole file for "v" and "f" entries.
     */
	while(!feof(file))
	{
		// read in the next line of the file
		fgets(buffer, sizeof(buffer), file);
		// and check for type of entry
		if(buffer[0] == 'v' && buffer[1] == ' ')
		{
				totalVertices++;

        }else if(buffer[0] == 'f' && buffer[1] == ' ')
        {
				totalFaces++;
		}
	}


    // now we know how many vertices and faces there are and may allocate memory
	vertices = new Vector4[totalVertices];
	polyList = new MyPolygon[totalFaces];

   
	// lets read in the actual data
	memset(buffer, '\0', sizeof(buffer));

	rewind(file);

	int vertexIndex = 0;
	int faceIndex = 0;
	char *token;
	char tempBuffer[256];
	char tempChar;
	char indexValue[5];
	int offset;
	
	while(!feof(file))
    {
		// read next line
		fscanf(file, "%c", &tempChar);

        if(tempChar == 'v')
        {
			
            fscanf(file, "%c", &tempChar);

            if(tempChar == ' ')
            {
				// read vertex coordinates
				fscanf(file, "%f", &vertices[vertexIndex].x);
                fscanf(file, "%f", &vertices[vertexIndex].y);
                fscanf(file, "%f", &vertices[vertexIndex].z);

                // read next line
                fscanf(file, "%c", &tempChar);

                vertexIndex++;
                }
		}
		else if(tempChar == 'f')
        {
			
			fscanf(file, "%c", &tempChar);
			
			if(tempChar == ' ')
			{
				// read rest of line
				fgets(buffer, sizeof(buffer), file);

				// copy to tempBuffer to work on it
				buffer[strlen(buffer)] = '\0';
				strcpy(tempBuffer, buffer);

				// set the search token
				token = strtok(tempBuffer, " ");

				// loop through the line to get the number spaces
				for (int count = 0; token != NULL; count++)
					token = strtok(NULL, " ");

				polyList[faceIndex].Indices = new int[count];
				polyList[faceIndex].numPoints = count;

				offset = 0;
				vertexIndex = 0;
				int bufferIndex = 0;

				for(int index = 0; index < polyList[faceIndex].numPoints; index++)
				{
					while(1)
                    {
						// if this index is not a space or newline then...
						if(buffer[bufferIndex] != ' ' && buffer[bufferIndex] != '\n')
                        {
							// ...we can increment the number of vertices in this line
                            vertexIndex++;
                        }
						else
                        {
					        // ...else we move to the next position in the buffer
                            bufferIndex++;
                            break;
                        }

							bufferIndex++;
						}

						for(int i = 0; i < vertexIndex; i++)
                        {
							// copy the entire number, offset is the number to add
							// to index to get the first char of this number, we add
							// to offset to get the whole number into indexValue.
							indexValue[i] = buffer[index + offset];
							offset++;
						}
						indexValue[vertexIndex] = '\0';
                     
						// save this vertex index value, if we get zero something is wrong
						// the indices in the file start from 1, not 0
						if(vertexIndex != 0)
							polyList[faceIndex].Indices[index] = atoi(&indexValue[0]);

						// instead of indices from 1 to n, we want them from 0 to n-1
						polyList[faceIndex].Indices[index] -= 1;
                     
						vertexIndex = 0;
					}

						faceIndex++;	
				}
		}
        else
		{
			// read end of line
			if(tempChar != '\n') fgets(buffer, sizeof(buffer), file);
		}

		memset(buffer, '0', sizeof (buffer));
	}
   
	fclose(file);


    // calculate normals
    Vector4 normal[3];

    for(int i = 0; i < totalFaces; i++)
	{
		normal[0] = vertices[polyList[i].Indices[0]];
        normal[1] = vertices[polyList[i].Indices[1]];
        normal[2] = vertices[polyList[i].Indices[2]];
        polyList[i].normal.FindTriangleNormal(normal);
	}

	return true;
}


void OBJLoader::clear()
{
	totalVertices = totalFaces = 0;
   
	if(vertices != NULL)
	{
		delete[] vertices;
        vertices = NULL;
    }

   
	if(polyList != NULL)
    {
		for(int i = 0; i < totalFaces; i++)
        {
			if(polyList[i].Indices != NULL)
			{
				delete[] polyList[i].Indices;
                polyList[i].Indices = NULL;
            }
        }
        delete[] polyList;
        polyList = NULL;
      
	}
}