// Showroom.cpp: Implementierung der Klasse Showroom
//
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <GL/wglext.h>
#include <corona.h>

#include "Showroom.hpp"
#include "Camera.h"
#include "BMPLoader.h"
#include "OBJLoader.h"
#include "RGBELoader.h"
#include "TGALoader.h"
#include "Matrix4.h"
#include "Vector4.h"
#include "Shader.h"
#include "Cg_Shader.h"
#include "FBufferObject.h"
#include "Wall.h"
#include "Water.h"
#include "Fire.h"

using namespace std;
using namespace corona;

/* define globals for movement keys */
#define up		0
#define down	1
#define left	2
#define right	3
#define rotl	4
#define rotr	5


//GLboolean keys[256];  // for 'more' keys :-)

#ifdef WIN32
#define M_PI 3.141592654
#define sinf(x) (float)sin((x))
#define cosf(x) (float)cos((x))
#define sqrtf(x) (float)sqrt((x))
#endif


#define d_camera 0.20 // amount to move camera by when using the keyboard arrow keys up/down
#define phi_camera (M_PI/180.0)*1.50 // radians to rotate camera by when using the keyboard arrow keys left/right

/*
 * some position constants
 */
#define floorY	-2.0
#define poolX	-10.0
#define poolY	(floorY + 1.0)
#define poolZ	 1
#define WaveInt	2


using std::ifstream;


/* vertex information for stored geometry */
typedef struct {
    GLfloat VA[3]; /* vertex coordinates */
    GLfloat NV[3]; /* vertex normal vector */
    GLfloat TV[3]; /* vertex tangent vector */
    GLfloat BV[3]; /* vertex binormal vector (N X T) */
    GLfloat TC[3]; /* texture coordinates */
} Vertex;


/* window dimensions */
int win_width = 1024;
int win_height = 768;
float win_ratio = 4.0f/3.0f;


/* key array for movement */
GLboolean key_array[6] = {GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE};

//GLboolean keys[256];	// for 'more' keys :-)

/* mouse position relative to upper left corner */
int mouse_x;
int mouse_y;
float mouse_sensivity = 512.0;

Camera Camera; // camera object used to move around

//OBJLoader OBJModel;
//BMPLoader BMPTexture;
TGALoader TGATexture;


// volumetric flame effect
Fire fire;

//time since the last update of the flame
float updateFlameTime = 1.0f;
//the maximum number of flame updates
const float flameUpdate = 80.0f;

// hdr texture image
//unsigned int hdrID = 0;
//int hdrWidth=0, hdrHeight=0;
//float exposure = 2.0f;

// hdr shader
//GLhandleARB toneShader;
//unsigned int glslToneTexture;
//unsigned int glslToneExposure;


// frame buffer
//unsigned int frameBufferIndex = 0;
//unsigned int depthBufferIndex = 0;
//unsigned int offScreenID = 0;

//bool shaders = false;
//bool bumpShaders = false;


// CG specific handles for bump mapping;
CGcontext bumpContext;
CGprogram bumpVertexProgram;
CGprofile bumpVertexProfile;
CGprogram bumpFragmentProgram;
CGprofile bumpFragmentProfile;

/* CG specific handles for wave simulation */
static CGcontext waterContext = NULL;
static CGprogram waterVertexProgram = NULL;
static CGprogram waterFragmentProgram = NULL;
static CGprofile waterVertexProfile;
static CGprofile waterFragmentProfile;
static CGparameter ModelViewProjParam = NULL;
static CGparameter timeParam = NULL;

/* booleans for toggling effects */
bool fire_on = true;
bool wallBump_on = true;
bool water_on = true;
bool waterBump_on = true;


char* gl_extensions;
char* extensions;

Wall wall_1(5.0f, 5.0f, 0.1f);
Wall wall_2(5.0f, 5.0f, 0.1f);
Wall wall_3(5.0f, 5.0f, 0.1f);
Wall door_1(5.0f, 5.0f, 0.1f);
Wall door_2(5.0f, 5.0f, 0.1f);
Wall floor_1(5.0f, 5.0f, 0.1f);

Water water(50,90,.1);
int ms = WaveInt;	// auxiliary function for the timer

Wall pool_wall_1(9.0f, 1.0f, 0.1f);
Wall pool_wall_2(5.0f, 1.0f, 0.1f);



/* pointer to screenshot data */
unsigned char* screenshotImage = 0;


Matrix4x4 ShadowMatrix;

Vector4 ground_normal(0.0f, 1.0f, 0.0f, 0.0f);

/* current light positions */
GLfloat light_direction_1[4] = {0.0f, -1.0f, 0.0f, 1.0f};
GLfloat light_position_1[4] = {0.0f, 3.0f, -1.0f, 1.0f};

GLfloat light_direction_2[4] = {0.0f, -1.0f, 0.0f, 1.0f};
GLfloat light_position_2[4] = {0.0f, 4.0f, 0.0f, 1.0f};

GLfloat light_direction_3[4] = {0.0f, -1.0f, 0.0f, 1.0f};
GLfloat light_position_3[4] = {-9.0f, 0.0f, 6.0f, 1.0f};

/* lighting parameters */

GLfloat ambient_light[4] = {0.0f, 0.0f, 0.5f, 1.0f }; 	
GLfloat diffuse_light[4] = {0.8f, 0.8f, 0.8f, 1.0f };
GLfloat specular_light[4] = {1.0f, 1.0f, 1.0f, 1.0f};
GLfloat specular_shininess = 1.0;



/*-------------------------------------------------------------------------------------------*/


/* 
 * callback function for redrawing
 */

void display_func(void)
{
	
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT|GL_ACCUM_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);

	// draw "light bulbs"
	/*
	glPushMatrix();
	glTranslatef(light_position_1[0], light_position_1[1], light_position_1[2]);
	glColor3f(1.0f, 1.0f, 1.0f);
	glutSolidSphere(0.1,10.0,10.0);
	glPopMatrix();
	*/

	/*
	glPushMatrix();
	glTranslatef(light_position_2[0], light_position_2[1], light_position_2[2]);
	glColor3f(1.0f, 1.0f, 1.0f);
	glutSolidSphere(0.1,10.0,10.0);
	glPopMatrix();
	*/

	glPushMatrix();
	glTranslatef(light_position_3[0], light_position_3[1], light_position_3[2]);
	glColor3f(1.0f, 1.0f, 1.0f);
	glutSolidSphere(0.1,10.0,10.0);
	glPopMatrix();


	// dras bumpmapped walls, door and floor
	glPushAttrib(GL_CURRENT_BIT|GL_LIGHTING_BIT|GL_TEXTURE_BIT);
	// disable OpenGL lighting since it is done by the shader
	glDisable(GL_LIGHTING);

	// enable the vertex and fragment profiles and bind the vertex and fragment programs
	
	if (wallBump_on) {
		cgGLEnableProfile(bumpVertexProfile);
		cgGLEnableProfile(bumpFragmentProfile);
		cgGLBindProgram(bumpVertexProgram);
		cgGLBindProgram(bumpFragmentProgram);
	}

	// get the parameters which we can pass to the vertex and fragment shaders
	CGparameter modelViewMatrix = cgGetNamedParameter(bumpVertexProgram, "modelViewProjMatrix");
	CGparameter lightPosition = cgGetNamedParameter(bumpVertexProgram, "vLightPosition");
	CGparameter cameraPosition = cgGetNamedParameter(bumpVertexProgram, "vCameraPosition");
	CGparameter lightDiffuseColor = cgGetNamedParameter(bumpFragmentProgram, "fLightDiffuseColor");
	CGparameter lightAmbientColor = cgGetNamedParameter(bumpFragmentProgram, "fLightAmbientColor");
	CGparameter lightSpecularColor = cgGetNamedParameter(bumpFragmentProgram, "fLightSpecularColor");
	CGparameter lightSpecularShininess = cgGetNamedParameter(bumpFragmentProgram, "fLightSpecularShininess");


	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, floor_1.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, floor_1.getBumpHandle());

	// set the light position parameter in the vertex shader
	cgGLSetParameter3f(lightPosition, light_position_3[0],light_position_3[1],light_position_3[2]);
	// set the camera position parameter in the vertex shader
	cgGLSetParameter3f(cameraPosition, Camera.xPos, Camera.yPos, Camera.zPos);
	// set the diffuse light in the fragment shader
	cgGLSetParameter3f(lightDiffuseColor, diffuse_light[0], diffuse_light[1], diffuse_light[2]);
	// set the ambient light in the fragment shader
	cgGLSetParameter3f(lightAmbientColor, ambient_light[0], ambient_light[1], ambient_light[2]);
	// set the specular light in the fragment shader
	cgGLSetParameter3f(lightSpecularColor, specular_light[0], specular_light[1], specular_light[2]);
	// ste the specular shininess in the fragment shader
	cgGLSetParameter1f(lightSpecularShininess, specular_shininess); 

	glMatrixMode(GL_MODELVIEW);
	// draw the floor
	glPushMatrix();
	glTranslatef(10.0f, floorY, 0.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();
	glPushMatrix();
	glTranslatef(10.0f, floorY, 5.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();
	glPushMatrix();
	glTranslatef(5.0f, floorY, 0.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();
	glPushMatrix();
	glTranslatef(5.0f, floorY, 5.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();
	glPushMatrix();
	glTranslatef(0.0f, floorY, 0.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();
	glPushMatrix();
	glTranslatef(0.0f, floorY, 5.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();
	glPushMatrix();
	glTranslatef(-5.0f, floorY, 0.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();
	glPushMatrix();
	glTranslatef(-5.0f, floorY, 5.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	glRotatef(-90.0, 1.0f, 0.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		floor_1.drawBump();
	else
		floor_1.draw();
	glPopMatrix();

	
	// draw the walls

	// draw wall_1
	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, wall_1.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, wall_1.getBumpHandle());	

	// left wall_1 on the left
	glPushMatrix();
	glTranslatef(10.0f, floorY, 0.0f);
	glRotatef(-90.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		wall_1.drawBump();
	else
		wall_1.draw();
	glPopMatrix();

	// draw wall_1 in the back
	glPushMatrix();
	glTranslatef(-5.0f, floorY, 10.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
		// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		wall_1.drawBump();
	else
		wall_1.draw();
	glPopMatrix();


	// draw wall_2
	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, wall_2.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, wall_2.getBumpHandle());	

	// draw wall_2 on the left
	glPushMatrix();
	glTranslatef(10.0f, floorY, 5.0f);
	glRotatef(-90.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		wall_2.drawBump();
	else
		wall_2.draw();
	glPopMatrix();

	// draw wall_2 on the right
	glPushMatrix();
	glTranslatef(-10.0f, floorY, 10.0f);
	glRotatef(90.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		wall_2.drawBump();
	else
		wall_2.draw();
	glPopMatrix();

	// draw wall_3
	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, wall_3.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, wall_3.getBumpHandle());	

	// draw wall_3 in the back
	glPushMatrix();
	glTranslatef(10.0f, floorY, 10.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		wall_3.drawBump();
	else
		wall_3.draw();
	glPopMatrix();

	// draw wall_3 on the right
	glPushMatrix();
	glTranslatef(-10.0f, floorY, 5.0f);
	glRotatef(90.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		wall_3.drawBump();
	else
		wall_3.draw();
	glPopMatrix();

	// draw door_1
	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, door_1.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, door_1.getBumpHandle());

	glPushMatrix();
	glTranslatef(5.0f, floorY, 10.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		door_1.drawBump();
	else
		door_1.draw();
	glPopMatrix();

	// draw door_2
	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, door_2.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, door_2.getBumpHandle());

	glPushMatrix();
	glTranslatef(0.0f, floorY, 10.0f);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		door_2.drawBump();
	else
		door_2.draw();
	glPopMatrix();


	// draw the pool walls

	// draw left border
	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, pool_wall_1.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, pool_wall_1.getBumpHandle());

	glPushMatrix();
	glTranslatef(-water.getSizeX() - water.getDx(), floorY, poolZ + water.getSizeY());
	glRotatef(90.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		pool_wall_1.drawBump();
	else
		pool_wall_1.draw();
	glPopMatrix();
	
	// draw near border
	// set the data for the first texturing unit
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, pool_wall_2.getDetailHandle());
	// set the data for the second texturing unit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, pool_wall_2.getBumpHandle());

	glPushMatrix();
	glTranslatef(-water.getSizeX() - water.getDx(), floorY, poolZ);
	glRotatef(180.0, 0.0f, 1.0f, 0.0f);
	// set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	if (wallBump_on)
		pool_wall_2.drawBump();
	else
		pool_wall_2.draw();
	glPopMatrix();


	// Disable the vertex and fragment profiles again
	if (waterBump_on) {
		cgGLDisableProfile(bumpVertexProfile);
		cgGLDisableProfile(bumpFragmentProfile);
	}

	
	/*
	 * Draw the water with the aid of shaders
	 */

	/* enable vertex and fragment profiles */
	if (water_on) {
		cgGLEnableProfile(waterVertexProfile);
		cgGLBindProgram(waterVertexProgram);
		if (waterBump_on) {
			cgGLEnableProfile(waterFragmentProfile);
			cgGLBindProgram(waterFragmentProgram);
		}
	}
	
	
	/* define the parameters for the vertex shader */
	CGparameter wModelViewProj = cgGetNamedParameter(waterVertexProgram, "ModelViewProj");
	CGparameter wLightPosition = cgGetNamedParameter(waterVertexProgram, "lightPos");
	CGparameter wCameraPosition = cgGetNamedParameter(waterVertexProgram, "camPos");
	
	// set the light position parameter in the vertex shader
	cgGLSetParameter3f(wLightPosition, light_position_3[0],light_position_3[1],light_position_3[2]);
	// set the camera position parameter in the vertex shader
	cgGLSetParameter3f(wCameraPosition, Camera.xPos, Camera.yPos, Camera.zPos);
	
	/* activate the appropriate texture (normal map) */
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);	
	glBindTexture(GL_TEXTURE_2D, water.getHandle());
	
	/* render the water surface */
	glPushMatrix();
	glTranslatef(poolX,poolY,poolZ);
	
	cgGLSetStateMatrixParameter(wModelViewProj, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	
	water.draw();
	
	glPopMatrix();

	/* disable the profiles again */
	
	if (water_on) {
		cgGLDisableProfile(waterVertexProfile);
		if (waterBump_on)
			cgGLDisableProfile(waterFragmentProfile);
	}
	
	glPopAttrib();

		
	// draw procedural fire effects
	if (fire_on) {
		glPushMatrix();
		//glScalef(0.5, 0.5, 0.5);	// scale fire volume
		glTranslatef(3.0f, -1.5f, 9.0f);
		//glScalef(2.0f, 3.0f, 2.0f);
		if (fire.bbox_enabled)
		{
			glColor3f(1, 1, 1);
			glutWireCube(1);
		}
		glTranslatef(-0.5f, -0.5f, -0.5f);
		if (fire.outlines_enabled)
			fire.draw_outlines();	// draw slice outlines
		if (fire.fire_enabled)
			fire.draw_fire(updateFlameTime);		// draw volumetric fire
		glPopMatrix();
	
		glPushMatrix();
		//glScalef(0.5, 0.5, 0.5);	// scale fire volume
		glTranslatef(-3.0f, -1.5f, 9.0f);
		//glScalef(2.0f, 3.0f, 2.0f);
		if (fire.bbox_enabled)
		{
			glColor3f(1, 1, 1);
			glutWireCube(1);
		}
		glTranslatef(-0.5f, -0.5f, -0.5f);
		if (fire.outlines_enabled)
			fire.draw_outlines();	// draw slice outlines
		if (fire.fire_enabled)
			fire.draw_fire(updateFlameTime);		// draw volumetric fire
		glPopMatrix();
	}


	// work in progress
	// Create the shadow matrix based on the surfaces' normal and the light position.
    /* ShadowMatrix.CreateShadowMatrix(ground_normal, Vector4(light_position_2[0],light_position_2[1],
															light_position_2[2],light_position_2[3]));

	// now we draw the shadow...
    glDisable(GL_DEPTH_TEST);

    glPushMatrix();
    //glMultMatrixf(ShadowMatrix.matrix);

    glBindTexture(GL_TEXTURE_2D, NULL);
	// black for shadow;
    glColor3f(0.0f, 0.0f, 0.0f);

    // and draw the objects which cast shadows onto the plane
    DrawCube();

	glPopMatrix();

    glEnable(GL_DEPTH_TEST);
	*/
	// ...end of building site


	
	// at the end swap buffers
	glutSwapBuffers();
}


/*
 * off-screen pass to render the hdr image
 */
 /*
void OffScreenPass()
{
	// Bind the frame buffer so we can render to it.
	glBindTexture(GL_TEXTURE_2D, 0);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, frameBufferIndex);

    // Clear the buffers and reset the model view matrix.
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

    // disable shaders no shaders
	 glUseProgramObjectARB(NULL);

    // bind the HDR texture image
    glEnable(GL_TEXTURE_RECTANGLE_NV);
    glBindTexture(GL_TEXTURE_RECTANGLE_NV, hdrID);

    // draw a simple quad object
    glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5f, -0.45f, -1.6f);
		glTexCoord2f(hdrWidth, 0.0f); glVertex3f(0.5f, -0.45f, -1.6f);
		glTexCoord2f(hdrWidth, hdrHeight); glVertex3f(0.5f, 0.45f, -1.6f);
        glTexCoord2f(0.0f, hdrHeight); glVertex3f(-0.5f, 0.45f, -1.6f);

    glEnd();

    // stop rendering to the frame buffer object
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}
*/

/*
 * now draw the hdr image
 */
/*
void TonePass()
{
	// Clear the buffers and reset the model view matrix.
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	// Prepare to draw full screen quad.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, win_width, 0, win_height);
   
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// enable shader
	glUseProgramObjectARB(toneShader);
	glUniform1iARB(glslToneTexture, 0);
	glUniform1fARB(glslToneExposure, exposure);

	// bind the full screen scene image
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, offScreenID);

   // draw the full screen quad   
	glBegin(GL_QUADS);

		glTexCoord2f(0, 0); glVertex2f(0, 0);
		glTexCoord2f(1, 0); glVertex2f(win_width, 0);
		glTexCoord2f(1, 1); glVertex2f(win_width, win_height);
		glTexCoord2f(0, 1); glVertex2f(0, win_height);

	glEnd();

	// return out of orthogonal mode
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}
*/

/* just a method to draw a snowman and play around
 * with it ;-)
 */
void DrawSnowman()
{
	// draw a snowman
	glTranslatef(2.5, -2.0, 4.0);
	glRotatef(180.0, 0.0, 1.0, 0.0);
	glColor3f(1.0f, 1.0f, 1.0f);

	// draw body	
	glTranslatef(0.0f ,0.75f, 0.0f);
	glutSolidSphere(0.75f,20,20);

	// draw head
	glTranslatef(0.0f, 1.0f, 0.0f);
	glutSolidSphere(0.25f,20,20);

	// draw eyes
	glPushMatrix();
	glColor3f(0.0f,0.0f,0.0f);
	glTranslatef(0.05f, 0.10f, 0.18f);
	glutSolidSphere(0.05f,10,10);
	glTranslatef(-0.1f, 0.0f, 0.0f);
	glutSolidSphere(0.05f,10,10);
	glPopMatrix();

	// draw nose
	glColor3f(1.0f, 0.5f , 0.5f);
	glRotatef(0.0f,1.0f, 0.0f, 0.0f);
	glutSolidCone(0.08f,0.5f,10,2);
}






/*
 * callback function for mouse events
 *
 * empty
 */

void mouse_func(int button, int state, int x, int y)
{
  
}


/*
 * callback function for mouse motion events
 * 
 * empty
 */
void motion_func(int x, int y)
{
    
}


/*
 * callback function for passive mouse motion events, e.g. w/o any buttons pressed
 * 
 */
void passive_motion_func(int x, int y)
{
	//printf("x-coord: %i ; ", x);
	//printf("y-coord: %i \n", y);

	// if the mouse was not moved simply return
	if((x == win_width/2) && (y == win_height/2))
		return;
	
	Camera.RotateByMouse(x, y, win_width/2.0, win_height/2.0);
	//printf("Camera was rotated by mouse");

	//reset the pointer to the middle of the screen
	glutWarpPointer(win_width/2.0f, win_height/2.0f);
}


/* 
 * this is the idle function where all updates are done for the simulation
 * loop if there is no user interaction
 *
 * empty
 */
void idle_func(void)
{

}


/*
 * callback function for reshape events
 */

void reshape_func(int new_width, int new_height)
{

	win_width  = new_width;
	win_height = new_height;
	win_ratio = (float)new_width/new_height;

	// reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0, 0, new_width, new_height);

	// set up the camera
	gluPerspective(45, win_ratio, 1, 1000);
	glMatrixMode(GL_MODELVIEW);
	updateView();

}



/*
 * callback functions for 'extended' keyboard keys
 * like cursor keys, page_up, page_down, insert, delete, F1..F12, and so on.
 */

void key_func_extended(int key, int x, int y)
{

	switch(key)
	{

    case GLUT_KEY_UP:
		key_array[up] = GL_TRUE;
        break;

    case GLUT_KEY_DOWN:
		key_array[down] = GL_TRUE;
        break;

	case GLUT_KEY_LEFT:
		if(glutGetModifiers()==GLUT_ACTIVE_SHIFT)
			key_array[left] = GL_TRUE;
		else
			key_array[rotl] = GL_TRUE;
		break;
	
	case GLUT_KEY_RIGHT:
		if(glutGetModifiers()==GLUT_ACTIVE_SHIFT)
			key_array[right]= GL_TRUE;
		else
			key_array[rotr] = GL_TRUE;
		break;
	}

}


/*
 * callback functions for 'extended' keyboard keys being released
 * like cursor keys, page_up, page_down, insert, delete, F1..F12, and so on.
 */	
void key_func_up_extended(int key, int x, int y)
{

/*	switch(key)
	{

    case GLUT_KEY_UP:
		key_array[up] = GL_FALSE;
        break;

    case GLUT_KEY_DOWN:
		key_array[down] = GL_FALSE;
        break;

	case GLUT_KEY_LEFT:
		if(glutGetModifiers()==GLUT_ACTIVE_SHIFT)
			key_array[left] = GL_FALSE;
		else
			key_array[rotl] = GL_FALSE;
		break;
	
	case GLUT_KEY_RIGHT:
		if(glutGetModifiers()==GLUT_ACTIVE_SHIFT)
			key_array[right]= GL_FALSE;
		else
			key_array[rotr] = GL_FALSE;
		break;
	}

*/
}

void updateView()
{
	glLoadIdentity();
    gluLookAt(Camera.xPos, Camera.yPos, Camera.zPos,
             Camera.xView, Camera.yView, Camera.zView,
             Camera.xUp, Camera.yUp, Camera.zUp);
	glutPostRedisplay();
}


/* 
 * timer callback function, has to be reregistered in each call;
 * basically this is the simulation loop, where all simulation updates
 * should be done on a periodic basis
 * 
 */

void timer_func(int value)
{
	
		
	if(key_array[up]){
		key_array[up] = GL_FALSE;
		Camera.MoveCamera(UP);
	}

	if(key_array[down]){
		key_array[down] = GL_FALSE;
		Camera.MoveCamera(DOWN);
	}

	if(key_array[left]){
		key_array[left] = GL_FALSE;
		Camera.StrafeCamera(STRAFE_LEFT);
		
	}

	if(key_array[right]){
		key_array[right] = GL_FALSE;
		Camera.StrafeCamera(STRAFE_RIGHT);
			
	}

	if(key_array[rotl]){
		key_array[rotl] = GL_FALSE;
		Camera.RotateCamera(LEFT,0.0,1.0,0.0);

	}

	if(key_array[rotr]){
		key_array[rotr] = GL_FALSE;
		Camera.RotateCamera(RIGHT,0.0,1.0,0.0);
		
	}

	//update particle systems

	//update object positions

	/* update the time parameter for the vertex
	 * shader being able to update the wave
	 * movement. */
	if (ms < WaveInt)
		ms++;
	else {
		float t;
		cgGLGetParameter1f(timeParam,&t);
		t += 1.0;
		cgGLSetParameter1d(timeParam,t);
		ms = 0.0;
	}

	// update procedural fire effect
	/*static unsigned int lastTime=0;
	unsigned int newTime;
	newTime = timeGetTime();
	updateFlameTime+=(float)(newTime-lastTime)*0.001f;
	lastTime=newTime;
	*/

	// update volumetric flame effect

	if (fire.animation)
	{
		fire.time += fire.speed;

		glh::ns_float::vec4 diffuse(0, 0, 0, 1);
		float k = frand();
		diffuse[0] = 0.5f + 0.3f * k;
		diffuse[1] = 0.1f + 0.2f * k;
		//glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse.v);
	}
	
	updateView();
	glutTimerFunc(10, timer_func, 1); // reregister the timer function

}



/*
 * initialize GL state machine, set up lighting, read in geometry, ...
 */
void initScene()
{

	/* set base color */
	glClearColor(0.0, 0.0, 0.0, 0.0);

	
	 /* turn on features */
	glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

	//glEnable(GL_STENCIL_TEST);

	/* select shade model; GL_FLAT or GL_SMOOTH */
	glShadeModel(GL_SMOOTH);

	/* sacrifice performance for better perspective look */
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    /* remove back faces to speed things up */
    //glCullFace(GL_BACK);

	/* set up coloring and lighting */
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular_light);
	glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 128);
	glEnable(GL_COLOR_MATERIAL);

	/*
	// first light source
	glLightfv(GL_LIGHT1, GL_AMBIENT, ambient_light);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse_light);
	glLightfv(GL_LIGHT1, GL_SPECULAR, specular_light);
	//glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 45.0f);
	//glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 75.0f);
	glLightfv(GL_LIGHT1, GL_POSITION, light_position_1);
	glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, light_direction_1);
	*/

	/*
    // second light source
    glLightfv(GL_LIGHT2, GL_AMBIENT, ambient_light);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, diffuse_light);
	glLightfv(GL_LIGHT2, GL_SPECULAR, specular_light);
	//glLightf(GL_LIGHT2, GL_SPOT_CUTOFF, 45.0f);
	//glLightf(GL_LIGHT2, GL_SPOT_EXPONENT, 75.0f);
	glLightfv(GL_LIGHT2, GL_POSITION, light_position_2);
	glLightfv(GL_LIGHT2, GL_SPOT_DIRECTION, light_direction_2);
	*/

	// third light source
    glLightfv(GL_LIGHT3, GL_AMBIENT, ambient_light);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, diffuse_light);
	glLightfv(GL_LIGHT3, GL_SPECULAR, specular_light);
	//glLightf(GL_LIGHT2, GL_SPOT_CUTOFF, 45.0f);
	//glLightf(GL_LIGHT2, GL_SPOT_EXPONENT, 75.0f);
	glLightfv(GL_LIGHT3, GL_POSITION, light_position_3);
	glLightfv(GL_LIGHT3, GL_SPOT_DIRECTION, light_direction_3);


	glEnable(GL_LIGHT1);
	glEnable(GL_LIGHT2);
	glEnable(GL_LIGHT3);
	glEnable(GL_LIGHTING);

	/* enable 2D texture mapping */
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_RECTANGLE_ARB);

	// read in textures
	loadTextures();

	// Cg procedural fire effect
	//fireTexture.InitCg();
	std::cout << "-intializing fire effect-\n";
	fire.init();
	std::cout << "-fire effect initialized-\n";

	// Cg bump mapping
	InitializeExtensions();
	//InitializeCg("Bumpmap.cg", bumpContext, bumpFragmentProfile, bumpFragmentProgram);
	InitializeCg("BumpVShader.cg", "BumpFShader.cg", bumpContext, bumpVertexProfile, bumpFragmentProfile, bumpVertexProgram, bumpFragmentProgram);
	
	InitializeCg("waterVShader.cg","waterFShader.cg",waterContext,waterVertexProfile, waterFragmentProfile, waterVertexProgram,waterFragmentProgram);
	timeParam = cgGetNamedParameter(waterVertexProgram, "time");

	// set initial camera parameters
	Camera.SetCamera(0.0f, 0.0f, -5.0f, 0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 0.0f);

}

/*
 * Load the textures
 */
void loadTextures()
{
	
	// load first wall texture
	wall_1.LoadDetail("wall_1.png");
	wall_1.LoadBumpmap("wall_1_bump.png");
	
	// load second wall texture
	wall_2.LoadDetail("wall_2.png");
	wall_2.LoadBumpmap("wall_2_bump.png");

	// load third wall texture
	wall_3.LoadDetail("wall_3.png");
	wall_3.LoadBumpmap("wall_3_bump.png");
	
	// load door textures
	door_1.LoadDetail("door_1.png");
	door_1.LoadBumpmap("door_1_bump.png");
	door_2.LoadDetail("door_2.png");
	door_2.LoadBumpmap("door_2_bump.png");
	
	// load floor texture
	floor_1.LoadDetail("floor_1.png");
	floor_1.LoadBumpmap("floor_1_normal.png");

	// load pool textures
	pool_wall_1.LoadDetail("pool_wall_1.png");
	pool_wall_1.LoadBumpmap("pool_wall_1_normal.png");
	pool_wall_2.LoadDetail("pool_wall_2.png");
	pool_wall_2.LoadBumpmap("pool_wall_2_normal.png");
	
	// load water textures
	water.loadTexture();

}

/*
 * Set parameters for the frame buffer
 */
/*
bool ConfigFBuffer()
{
	// generate frame buffer object and bind it
	glGenFramebuffersEXT(1, &frameBufferIndex);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, frameBufferIndex);
	glGenRenderbuffersEXT(1, &depthBufferIndex);

	// create texture to render to
	glGenTextures(1, &offScreenID);
	glBindTexture(GL_TEXTURE_2D, offScreenID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, win_width, win_height, 0,
                GL_RGB, GL_FLOAT, 0);

	// bind texture to frame buffer
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                             GL_TEXTURE_2D, offScreenID, 0);

	// initialize the render buffer
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depthBufferIndex);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, win_width, win_height);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
                                GL_RENDERBUFFER_EXT, depthBufferIndex);

	// check for errors
	if(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT)
		return false;

	// return output of frame buffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	return true;
}
*/


/*
 * callback function for keyboard events
 */

void key_func(unsigned char key, int x, int y)
{
    switch(key)
    {
    case 'b':
	case 'B':
		if (wallBump_on)
			wallBump_on = false;
		else
			wallBump_on = true;
	break;
	case '\033':
		exit(0);
	break;
	case 'F':
	case 'f':
		if (fire_on)
			fire_on = false;
		else
			fire_on = true;
	break;
	case 'g':
		if(glutGameModeGet(GLUT_GAME_MODE_ACTIVE)) 
		{
			glutLeaveGameMode(); //crashes at the moment
		}
	break;
	case 'G':
		glutGameModeString("1280x1024:32");

		// enter full screen
		if(glutGameModeGet(GLUT_GAME_MODE_POSSIBLE))
		{
			glutEnterGameMode();
			//callbacks have to be registered again since technically speaking we have a new context
			/* register callback functions */
			glutDisplayFunc(display_func);
			//glutIgnoreKeyRepeat(1);
			glutKeyboardFunc(key_func);
			//glutMouseFunc(mouse_func);
			//glutMotionFunc(motion_func);
			glutPassiveMotionFunc(passive_motion_func);
			//glutIdleFunc(idle_func);
			glutReshapeFunc(reshape_func);
			glutSpecialFunc(key_func_extended); // callback to handle "extended" keyboard input
			glutSpecialUpFunc(key_func_up_extended);
			glutTimerFunc(1, timer_func, 1); // initially register timer function 
		}else{
			printf("The selected mode is not available\n");
		}
	break;
	// save a screenshot to a .tga file
	case 'S':
    case 's':
        // allocate neccessary memory
        screenshotImage = new unsigned char[win_width * win_height * 3];
		// use glReadPixels() to read every pixel on the screen
		// that one specifies; must use one less than each size
		glReadPixels(0, 0, win_width, win_height, GL_RGB, GL_UNSIGNED_BYTE, screenshotImage);
        // all our WriteTGA() function to write out the tga
        TGATexture.WriteTGA("Output.tga", win_width, win_height, screenshotImage);
        delete[] screenshotImage;
	break;
	case 'W':
	case 'w':
		if (water_on)
			water_on = false;
		else
			water_on = true;
	break;
	case 'V':
	case 'v':
		if (waterBump_on)
			waterBump_on = false;
		else
			waterBump_on = true;
	break;


	// change the exposure for hdr rendering
	/*case GLUT_KEY_PAGE_UP:
		exposure += 0.05f;
	break;
	case GLUT_KEY_PAGE_DOWN:
		exposure -= 0.05f;
		if(exposure < 0.0f)
			exposure = 0.01f;
	break;
	*/

    case '?':
	case 'h':
    default:
	fprintf(stderr, "Keyboard commands:\n"
		"b or B - toggle bumpmapping on walls\n"
		"f or F - toggle fire\n"
		"g - leave glut game mode (crashes at the moment!)\n"
		"G - enter glut game mode\n"
		"h or ? - print this manual\n"
	    "s or S - take a screenshot\n"
		"w or W - toggle wave simulation\n"
		"v or V - toggel bumpmapping on water\n");
		//"Page_up: Raise HDR exposure\n"
		//"Page_down: Lower HDR exposure\n");
	break;
    }

}

void clearCGContext(CGcontext context)
{
	// destroying the CG context automatically destroys all attached CG programs
	cgDestroyContext(context);
}

	
main(int argc, char *argv[])
{

	std::cout << "Initializing glut...";
    glutInit(&argc, argv);
    glutInitWindowSize(win_width, win_height);
	glutInitWindowPosition(100, 100);
    glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE|GLUT_DEPTH);//|GLUT_STENCIL|GLUT_ACCUM);
    glutCreateWindow("Showroom");
	std::cout << "...done\n.";
   
  
    /* initialize sceen */
	initScene();

	/* register callback functions */
    glutDisplayFunc(display_func);
	//glutIgnoreKeyRepeat(1);
    glutKeyboardFunc(key_func);
    //glutMouseFunc(mouse_func); // for the mouse buttons
    //glutMotionFunc(motion_func); // with mouse buttons pressed
	glutPassiveMotionFunc(passive_motion_func); // w/o any mouse buttons pressed
	//glutIdleFunc(idle_func);
    glutReshapeFunc(reshape_func);
	glutSpecialFunc(key_func_extended); // callback to handle "extended" keyboard input
	//glutSpecialUpFunc(key_func_up_extended); // callback to handle "extended" up key events
	glutTimerFunc(1, timer_func, 1); // initially register timer function

    glutMainLoop();

	cgDestroyContext(waterContext);

	return(0);

}