#include "Water.h"



	/*
	 * Default constructor.
	 */
	Water::Water() : nX(100), nY(100), dx((float)1e-1)  {

		verts = (Vector4 *) malloc(nX*nY*sizeof(Vector4));
		initGrid(nX,nY);
	}

	
	/*
	 * parametrized constructor
	 *
	 * int _nX		- number of vertices on x axis
	 * int _nY		- number of vertices on y axis
	 * float _dx	- distance between to adjacent vertices
	 */
	Water::Water(int _nX, int _nY, float _dx) : nX(_nX), nY(_nY), dx(_dx) {
		verts = (Vector4 *) malloc(nX*nY*sizeof(Vector4));

		initGrid(nX,nY);
		

	}

	
	/*
	 * initializes the coordinates of the vertices
	 *
	 * int nX	- number of vertices on x axis
	 * int nY	- number of vertices on y axis
	 */
	void Water::initGrid(int nX,int nY) {
		for (int x = 0; x < nX; x++) {
			float d = x*dx;
			for (int y = 0; y < nY; y++) {
				verts[x*nY + y].x = d;
				verts[x*nY + y].y = -0.5;
				verts[x*nY + y].z = y*dx;
			}
		}
	}


	/*
	 * copy constructor
	 */
	Water::Water(const Water &w) : nX(w.nX), nY(w.nX), dx(w.dx), verts(w.verts) {}

	
	/*
	 * renders a triangle mesh to the scene and links them to
	 * their texture positions
	 */
	Water::draw() {
		for (int x = 0; x < nX-1; x++) {
			glBegin(GL_TRIANGLE_STRIP);
			for (int y = 0; y < nY-1; y++) {
				// 'even' vertex
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, verts[x*nY+y].x/getSizeX(), verts[x*nY+y].z/getSizeY());
				glVertex3f(verts[x*nY+y].x,verts[x*nY+y].y,verts[x*nY+y].z);
				
				// 'odd' vertex
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, verts[(x+1)*nY+y].x/getSizeX(), verts[(x+1)*nY+y].z/getSizeY());
				glVertex3f(verts[(x+1)*nY+y].x,verts[(x+1)*nY+y].y,verts[(x+1)*nY+y].z);
			}
			glEnd();
		}
	}


	/*
	 * loads the normal map and binds it to an appropriate
	 * handler.
	 */
	Water:: loadTexture() {
		Image* waterTexture = FlipImage(OpenImage("water_bump.nm",PF_R8G8B8),CA_X);
		Array3<unsigned char> waterData(waterTexture->getWidth(), waterTexture->getHeight(),3,(unsigned char*)waterTexture->getPixels());

		glPushAttrib(GL_TEXTURE_BIT);

		// Generate the texture
		glGenTextures(1, &waterHandle);
		glBindTexture(GL_TEXTURE_2D, waterHandle);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);

		gluBuild2DMipmaps(GL_TEXTURE_2D,
			GL_RGB8,
			waterData.getWidth(),
			waterData.getHeight(),
			GL_RGB,
			GL_UNSIGNED_BYTE,
			waterData.getData());
		
		glPopAttrib();
}


	/*
	 * some accessors which might be useful
	 */

	float Water::getSizeX() { return nX*dx;}

	float Water::getSizeY() { return nY*dx;}

	float Water::getDx() { return dx;}

	unsigned int Water::getHandle() { return waterHandle;}

	Vector4 *Water::getVerts() { return verts;}
